package de.ozgcloud.admin.common.errorhandling;

import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import java.util.Set;
import java.util.stream.Stream;

import jakarta.validation.ConstraintViolation;
import jakarta.validation.ConstraintViolationException;
import jakarta.validation.Validation;
import jakarta.validation.Validator;
import jakarta.validation.constraints.NotEmpty;

import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.http.HttpStatus;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;

import de.ozgcloud.admin.RootController;
import de.ozgcloud.admin.RootModelAssembler;
import de.ozgcloud.common.test.ITCase;
import lombok.Builder;
import lombok.Getter;
import lombok.SneakyThrows;

@ITCase
@AutoConfigureMockMvc
@WithMockUser
class ExceptionControllerITCase {

	@Autowired
	private MockMvc mockMvc;

	@MockBean
	private RootModelAssembler modelAssembler;

	@Nested
	class TestExceptions {
		@ParameterizedTest
		@MethodSource("exceptionAndExpectedStatus")
		@SneakyThrows
		void shouldHandleExceptionWithStatus(Class<? extends Exception> exceptionClass, HttpStatus expectedStatus) {
			when(modelAssembler.toModel(any())).thenThrow(TestErrorController.EXCEPTION_PRODUCER.get(exceptionClass).produceException());

			var result = performGet();

			result.andExpect(status().is(expectedStatus.value()));
		}

		@ParameterizedTest
		@MethodSource("exceptionAndExpectedStatus")
		@SneakyThrows
		void shouldRespondWithStatusInBody(Class<? extends Exception> exceptionClass, HttpStatus expectedStatus) {
			when(modelAssembler.toModel(any())).thenThrow(exceptionClass);

			var result = performGet();

			result.andExpect(jsonPath("$.status").value(expectedStatus.value()));
		}

		private static Stream<Arguments> exceptionAndExpectedStatus() {
			return ExceptionController.STATUS_BY_EXCEPTION.entrySet().stream().map(kv -> Arguments.of(kv.getKey(), kv.getValue()));
		}

	}

	@Nested
	class TestConstraintViolationException {
		@Test
		@SneakyThrows
		void shouldHaveValidationMessage() {
			when(modelAssembler.toModel(any())).thenAnswer((a) -> {
				throw new ConstraintViolationException(getConstraintViolations());
			});

			var result = performGet();

			result.andExpect(jsonPath("$.detail").value("string: Empty field"));
		}

		private Set<ConstraintViolation<ValidatedClass>> getConstraintViolations() {
			Validator validator = Validation.buildDefaultValidatorFactory().getValidator();
			return validator.validate(ValidatedClass.builder().build());
		}

		@Getter
		@Builder
		private static class ValidatedClass {
			@NotEmpty(message = "Empty field")
			private String string;
		}

	}

	@SneakyThrows
	private ResultActions performGet() {
		return mockMvc.perform(get(RootController.PATH));
	}

}
