/*
 * Copyright (c) 2024. Das Land Schleswig-Holstein vertreten durch das Ministerium für Energiewende, Klimaschutz, Umwelt und Natur
 * Zentrales IT-Management
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.admin.common.errorhandling;

import static de.ozgcloud.admin.common.errorhandling.ExceptionController.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.springframework.http.HttpStatus;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;

import lombok.SneakyThrows;

class ExceptionControllerTest {

	private MockMvc mockMvc;

	@BeforeEach
	void setup() {
		mockMvc = MockMvcBuilders
				.standaloneSetup(new TestErrorController())
				.setControllerAdvice(new ExceptionController()).build();
	}

	@DisplayName("Error handler")
	@Nested
	class TestErrorHandler {

		@ParameterizedTest
		@MethodSource("exceptionAndExpectedStatus")
		@SneakyThrows
		void shouldHandleExceptionWithStatus(Class<? extends Exception> exceptionClass, HttpStatus expectedStatus) {
			var result = doPerformWithError(exceptionClass);

			result.andExpect(status().is(expectedStatus.value()));
		}

		@ParameterizedTest
		@MethodSource("exceptionAndExpectedStatus")
		@SneakyThrows
		void shouldRespondWithStatusInBody(Class<? extends Exception> exceptionClass, HttpStatus expectedStatus) {
			var result = doPerformWithError(exceptionClass);

			result.andExpect(jsonPath("$.status").value(expectedStatus.value()));
		}

		@ParameterizedTest
		@MethodSource("exceptionAndExpectedStatus")
		@SneakyThrows
		void shouldRespondWithTitler(Class<? extends Exception> exceptionClass, HttpStatus expectedStatus) {
			var result = doPerformWithError(exceptionClass);

			result.andExpect(jsonPath("$.title").exists());
		}

		@ParameterizedTest
		@MethodSource("exceptionAndExpectedStatus")
		@SneakyThrows
		void shouldRespondWithDetail(Class<? extends Exception> exceptionClass, HttpStatus expectedStatus) {
			var result = doPerformWithError(exceptionClass);

			result.andExpect(jsonPath("$.detail").exists());
		}

		@ParameterizedTest
		@MethodSource("exceptionAndExpectedStatus")
		@SneakyThrows
		void shouldRespondWithInstance(Class<? extends Exception> exceptionClass, HttpStatus expectedStatus) {
			var result = doPerformWithError(exceptionClass);

			result.andExpect(jsonPath("$.instance").exists());
		}

		@ParameterizedTest
		@MethodSource("exceptionAndExpectedStatus")
		@SneakyThrows
		void shouldRespondWithType(Class<? extends Exception> exceptionClass, HttpStatus expectedStatus) {
			var result = doPerformWithError(exceptionClass);

			result.andExpect(jsonPath("$.type").exists());
		}

		private static Stream<Arguments> exceptionAndExpectedStatus() {
			return STATUS_BY_EXCEPTION.entrySet().stream().map(kv -> Arguments.of(kv.getKey(), kv.getValue()));
		}

		@SneakyThrows
		private ResultActions doPerformWithError(Class<? extends Exception> exceptionClass) {
			return mockMvc.perform(get("/api/test-error").param("errorClassName", exceptionClass.getName()));
		}
	}

	@DisplayName("ResourceNotFound error")
	@Nested
	class TestResourceNotFoundError {

		@Test
		@SneakyThrows
		void shouldHaveStatus() {
			var result = doRequestUnknown();

			result.andExpect(status().is(HttpStatus.NOT_FOUND.value()));
		}

		@Test
		@SneakyThrows
		void shouldRespondWithStatusInBody() {
			var result = doRequestUnknown();

			result.andExpect(jsonPath("$.status").value(HttpStatus.NOT_FOUND.value()));
		}

		@SneakyThrows
		private ResultActions doRequestUnknown() {
			return mockMvc.perform(get("/api/unknown"));
		}
	}
}
